﻿//*************************************************************************************************
//【サンプルプログラム】『Kinectからカラー画像を取得し、人物部分だけを切り取って表示する』
//【名称】Bitmap_GetPlayerColorImage　　【Version】1.02
//【作成者】私設研究所Neo-Tech-Lab.com  上田智章
//【作成日】2012年3月5日　【コメント追記】2012年6月30日
//【注意事項】
//　このプログラムをKinect for Windows SDK Version1.5で動かすには、参照設定の
// Microsoft.Kinectを一旦削除し、プロジェクト(P)-参照の追加(F)で、.NETタブ内
// にあるMicrosoft.Kinectを選択・追加してください。 
//*************************************************************************************************

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Diagnostics;             //【Debug】
using System.Drawing;                 //【Bitmap】【Color】
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;         //【BitmapData】
using System.Runtime.InteropServices; //【Marshal.Copy】
using Microsoft.Kinect;               //【参照設定】プロジェクト(P)⇒参照の追加(F)⇒.NETタブ⇒Microsoft.Kinectを選択

namespace Bitmap_GetPlayerColorImage
{
    public partial class Form1 : Form
    {
        //【KINECT関連】
        //　関数内でダイナミックに変数を確保すると動作速度が遅くなるので、ここにまとめて変数を確保しています。
        private const int KS_FrameWidth = 640;  //【KINECTのFrameの横方向画素数】
        private const int KS_FrameHeight = 480; //【KINECTのFrameの縦方向画素数】
        private KinectSensor KS_KinectSensor;   //【KINECT】
        private byte[] KS_ColorImageDataXRGB;   //【KINECTのColorImageData】32bit/画素,X8R8G8B8表示形式
        private Bitmap KS_bmp1;                 //【ColorImage表示用Bitmapオブジェクト】
        private BitmapData KS_data1;            //【ColorImage用BitmapData】
        private Rectangle KS_rect1;             //【ColorImage表示枠】
        private DepthImagePixel[] KS_DepthData; //【KINECTのDepthImageData】16bit/画素(MSB13bit:Depth, LSB3bit:PlayerID) 
        private ColorImagePoint[] KS_MapData;   //【Coordinate Mapper】
        private int[] KS_MapIndex;              //【Coordinate Mapping Index】
        private int KS_Lmin;                    //【深度測定可能最近距離】default:  800mm, near: 400mm 
        private int KS_Lmax;                    //【深度測定可能最遠距離】default: 4000mm, near:3000mm
        private bool KS_flagFlip = true;        //【左右反転フラグ】
        //【NTL拡張現実センサ関連】---------------------------------------------------------
        private NTL_ColorBar ColorBar = new NTL_ColorBar(); //【デプス値をカラーに変換する部品】
        private int Count;

        public Form1()
        {
            InitializeComponent();

            //***********************************************************
            //【Kinect検出】【共通】
            //***********************************************************
            //【利用可能なKINECTセンサを取得する】
            KS_KinectSensor = (from sensorToCheck
                               in KinectSensor.KinectSensors
                               where sensorToCheck.Status == KinectStatus.Connected
                               select sensorToCheck).FirstOrDefault();
            //【検出結果の表示】
            pictureBox2.SizeMode = PictureBoxSizeMode.StretchImage;
            if (KS_KinectSensor == null) //【Kinect接続が確認できなかった場合】
            {
                pictureBox2.Image = Bitmap.FromFile("../../image/KinectOff.png");
                MessageBox.Show(
                   "接続可能なKINECTセンサが見当たりません。",
                   "KINECT for Windows Error Message",
                   MessageBoxButtons.OK
                );
                return;
            }
            pictureBox2.Image = Bitmap.FromFile("../../image/KinectOn.png");

            //***********************************************************
            //【Kinectの初期化処理】
            //***********************************************************
            //【KINECTの設定】ColorStream, DepthStream, SkeletonStreamを許可
            KS_KinectSensor.ColorStream.Enable(ColorImageFormat.RgbResolution640x480Fps30); //【ColorStream許可】
            KS_KinectSensor.DepthStream.Enable(DepthImageFormat.Resolution640x480Fps30);    //【DepthStream許可】
            KS_KinectSensor.SkeletonStream.Enable();                                        //【SkeltonStream許可】
            //【★★★デプス情報のPlayerIDが有効になるのはSkeletonStreamを有効にしている場合に限られるので注意★★★】
            //【KINECT動作開始】
            KS_KinectSensor.Start(); //【KINECTスタート】

            //【KINECTデータ取得イベントの設定とデータバッファの確保】
            KS_KinectSensor.ColorFrameReady += ColorImageReady; //【準備完了でColorImageReadyがcallされる】
            KS_KinectSensor.DepthFrameReady += DepthImageReady; //【準備完了でDepthImageReadyがcallされる】
            KS_ColorImageDataXRGB = new byte[4 * KS_FrameWidth * KS_FrameHeight]; //【カラー画像データ用】
            KS_bmp1 = new Bitmap(KS_FrameWidth, KS_FrameHeight, System.Drawing.Imaging.PixelFormat.Format32bppRgb); //【ColorImage表示用】
            KS_rect1 = new Rectangle(0, 0, KS_FrameWidth, KS_FrameHeight); //【ColorImage用枠】
            KS_DepthData = new DepthImagePixel[KS_FrameWidth * KS_FrameHeight]; //【深度情報データ用】
            KS_MapData = new ColorImagePoint[KS_FrameWidth * KS_FrameHeight];
            KS_MapIndex = new int[KS_FrameWidth * KS_FrameHeight];
            Count = 0;

            KS_KinectSensor.DepthStream.Range = DepthRange.Default;    //【Depth測定モードをDefaultに設定】
            KS_KinectSensor.CoordinateMapper.MapDepthFrameToColorFrame(
                DepthImageFormat.Resolution640x480Fps30,
                KS_DepthData,
                ColorImageFormat.RgbResolution640x480Fps30,
                KS_MapData);
            for (int i = 0; i < KS_MapData.Length; i++)
            {
                KS_MapIndex[i] = 4*(KS_MapData[i].X + KS_FrameWidth * KS_MapData[i].Y);
            }
            KS_CheckDistance();                                        //【設定結果を表示】Lmax, Lminを取得
            pictureBox1.SizeMode = PictureBoxSizeMode.StretchImage; //【キャプチャー画像がpictureBox枠内にはめ込まれます】

            //【Kinectヘッドの仰角を設定】設定範囲は±27°
            int Angle = 0;
            KS_KinectSensor.ElevationAngle = Angle;
            textBox1.Text = Angle.ToString();

            ColorBar.CreateColorPalletForAbsolute(KS_Lmax + 1);     //【カラーパレットを作成する】
            pictureBox3.SizeMode = PictureBoxSizeMode.StretchImage;
            pictureBox3.Image = ColorBar.CreateColorScale(256, 10); //【256×10画素サイズのカラーバーを作成して表示する】

            this.FormClosing += new FormClosingEventHandler(Form1_FormClosing);
            radioButton1.CheckedChanged += new EventHandler(DepthRangeChange);
        }

        //***************************************************************
        //【終了処理】
        //***************************************************************
        void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (KS_KinectSensor != null)
            {
                KS_KinectSensor.Stop();    //【KINECTセンサを終了】★これを実行しないとレーザーが消えない
                KS_KinectSensor.Dispose(); //【KINECTセンサのオブジェクトを解放する】
            }
        }

        //***************************************************************
        //【距離モード変更】Defaultモード(800mm～4000mm) Nearモード(400m～3000mm)
        //***************************************************************
        void DepthRangeChange(object sender, EventArgs e)
        {
            if (radioButton1.Checked) { KS_KinectSensor.DepthStream.Range = DepthRange.Default; label8.Text = "Default"; }
            else { KS_KinectSensor.DepthStream.Range = DepthRange.Near; label8.Text = "Near"; }
            KS_CheckDistance();
        }

        //***************************************************************
        //【CheckDistance】KINECTセンサの検出可能距離現在値を調べる
        //***************************************************************
        private void KS_CheckDistance()
        {
            KS_Lmax = KS_KinectSensor.DepthStream.MaxDepth;
            KS_Lmin = KS_KinectSensor.DepthStream.MinDepth;
            label1.Text = "Min=" + KS_Lmin.ToString() + "mm";
            label2.Text = "Max=" + KS_Lmax.ToString() + "mm";
        }

        //***************************************************************
        //【上下角変更ボタン】KINECTの上下角を変更。±27°の範囲内。時間を置かないとエラーが発生する
        //***************************************************************
        private void button2_Click(object sender, EventArgs e)
        {
            Int32 angle = int.Parse(textBox1.Text); //【設定角度を読み込む】±27°
            if (angle <= KS_KinectSensor.MaxElevationAngle && angle >= KS_KinectSensor.MinElevationAngle)
            {
                KS_KinectSensor.ElevationAngle = angle; //【上下角の変更】
            }
        }

        //***************************************************************
        //【カラー画像の左右反転指定フラグ】
        //***************************************************************
        private void checkBox1_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBox1.Checked) { KS_flagFlip = true; } else { KS_flagFlip = false; }
        }

        //***************************************************************
        //【ColorImageの取得イベント】
        //***************************************************************
        private void ColorImageReady(object sender, ColorImageFrameReadyEventArgs e)
        {
            using (ColorImageFrame f = e.OpenColorImageFrame())
            {
                if ((Count>5) && (f != null))
                {
                    f.CopyPixelDataTo(KS_ColorImageDataXRGB); //【KINECTのColor画像を配列PixelDataにコピー】
                    KS_ConvertDepthToColor();

                    KS_data1 = KS_bmp1.LockBits(KS_rect1, ImageLockMode.WriteOnly, System.Drawing.Imaging.PixelFormat.Format32bppRgb); //【Lock】
                    Marshal.Copy(KS_ColorImageDataXRGB, 0, KS_data1.Scan0, KS_ColorImageDataXRGB.Length); //【UnmanagedMemoryのBitmapDataに配列PixelDataをコピー】
                    KS_bmp1.UnlockBits(KS_data1);           //【Lock解除】
                    if (KS_flagFlip) { KS_bmp1.RotateFlip(RotateFlipType.Rotate180FlipY); } //【左右反転】
                    pictureBox1.Image = KS_bmp1;            //【pictureBox1に画像表示】
                }
                else
                {
                    Count++;
                    //【ImageFrameがnullだったのは要求が間に合わなかったと考えられます】
                }
            }
        }

        //***************************************************************
        //【DepthImageの取得イベント】
        //***************************************************************
        private void DepthImageReady(object sender, DepthImageFrameReadyEventArgs e)
        {
            using (DepthImageFrame f = e.OpenDepthImageFrame())
            {
                if (f != null)
                {
                    f.CopyDepthImagePixelDataTo(KS_DepthData); //【KINECTのDepth情報を配列DepthDataにコピー】
                }
                else
                {
                    //【DepthFrameがnullだったのは要求が間に合わなかったと考えられます】
                }
            }
        }

        //***************************************************************
        //【ConvertDepthToColor】16ビットのDepth値をカラーバーに対応する色に変更
        //***************************************************************
        private void KS_ConvertDepthToColor()
        {
            int i, j, d;  //【Nearモード】  400～3000mm ⇒ 0.0980769　2600色
            for (i = 0; i < KS_DepthData.Length; i++)
            {
                if (KS_DepthData[i].PlayerIndex > 0)
                {
                    j = KS_MapIndex[i];
                    d = 4 * KS_DepthData[i].Depth;            //【距離情報】
                    KS_ColorImageDataXRGB[j] = ColorBar.NTL_ColorElement[d]; //【青】B
                    KS_ColorImageDataXRGB[j + 1] = ColorBar.NTL_ColorElement[d + 1]; //【緑】G
                    KS_ColorImageDataXRGB[j + 2] = ColorBar.NTL_ColorElement[d + 2]; //【赤】R
                    KS_ColorImageDataXRGB[j + 3] = ColorBar.NTL_ColorElement[d + 3]; //【α】

                }
            }
        }
    }
}
